module mnozenie(
    input   logic          rst,
    input   logic          clk,
    input   logic [ 511:0] a,
    input   logic [ 511:0] b,
    input   logic          enable,
    output  logic          result_valid,
    output  logic [1023:0] result
);

//====================================================
//=======  FSM Parameters  ===========================
//====================================================
typedef enum logic [2:0] {
    IDLE              = 3'b000,
    MULT_STATE        = 3'b001,
    SUM_STATE1        = 3'b010,
    SUM_STATE2        = 3'b011,
    RESULT_STATE      = 3'b100
} main_fsm_t;

main_fsm_t main_fsm;

logic   [255:0]        al;
logic   [255:0]        ah;
logic   [255:0]        bl;
logic   [255:0]        bh;
assign al = a[255:  0];
assign ah = a[511:256];
assign bl = b[255:  0];
assign bh = b[511:256];

logic [ 511:0] al_bl;
logic          al_bl_valid;
logic [ 511:0] ah_bl;
logic          ah_bl_valid;
logic [ 511:0] al_bh;
logic          al_bh_valid;
logic [ 511:0] ah_bh;

logic   [1023:0]        shl_al_bl       ;//al*bl     
logic   [1023:0]        shl_ah_bl       ;//ah*bl << 256
logic   [1023:0]        shl_al_bh       ;//al*bh << 256
logic   [1023:0]        shl_ah_bh       ;//ah*bh << 512

logic   [1023:0]        sum1            ;//al*bl + (ah*bl << 256)
logic   [1023:0]        sum2            ;//(al*bh  << 256) + (ah*bh << 512)

logic                       mult_enable     ;
logic                       sum1_2_enable   ;
logic                       sum3_enable     ;
logic                       sum1_valid      ;
logic                       sum2_valid      ;
logic                       sum1_clear      ;
logic                       sum2_clear      ;
logic                       sum3_clear      ;

assign shl_al_bl = {512'b0, al_bl};
assign shl_ah_bl = {256'b0, ah_bl, 256'b0};
assign shl_al_bh = {256'b0, al_bh, 256'b0};
assign shl_ah_bh = {ah_bh, 512'b0};

//====================================================
//=======  INSTANTIONS  ==============================
//====================================================
mult mult_inst1 (
    .clock          ( clk ),
    .reset          ( rst ),
    .enable         ( mult_enable ),
    .dataa          ( al ),
    .datab          ( bl ),
    .result         ( al_bl ),
    .result_valid   ( al_bl_valid )
);

mult mult_inst2 (
    .clock          ( clk ),
    .reset          ( rst ),
    .enable         ( mult_enable ),
    .dataa          ( al ),
    .datab          ( bh ),
    .result         ( al_bh ),
    .result_valid   ( al_bh_valid )
);

mult mult_inst3 (
    .clock          ( clk ),
    .reset          ( rst ),
    .enable         ( mult_enable ),
    .dataa          ( ah ),
    .datab          ( bl ),
    .result         ( ah_bl ),
    .result_valid   ( ah_bl_valid )
);

mult    mult_inst4 (
    .clock          ( clk ),
    .reset          ( rst ),
    .enable         ( mult_enable ),
    .dataa          ( ah ),
    .datab          (  bh ),
    .result         ( ah_bh ),
    .result_valid   ( ah_bh_valid )
);

sumator u_sumator1(
    .rst    (rst),
    .clk    (clk),
    .a      (shl_al_bl),
    .b      (shl_ah_bl),
    .enable (sum1_2_enable),
    .clear  (sum1_clear),
    .c_valid (sum1_valid),
    .c      (sum1)
);

sumator u_sumator2(
    .rst    (rst),
    .clk    (clk),
    .a      (shl_al_bh),
    .b      (shl_ah_bh),
    .enable (sum1_2_enable),
    .clear  (sum2_clear),
    .c_valid (sum2_valid),
    .c      (sum2)
);

sumator u_sumator3(
    .rst    (rst),
    .clk    (clk),
    .a      (sum1),
    .b      (sum2),
    .enable (sum3_enable),
    .clear  (sum3_clear),
    .c_valid (result_valid),
    .c      (result)
);

//====================================================
//=======  Main FSM  ==================
//====================================================
always_ff @(posedge clk or posedge rst) begin

    if (rst) begin
        main_fsm <= IDLE;

        mult_enable   <= 1'b0;
        sum1_2_enable <= 1'b0;
        sum3_enable   <= 1'b0;
        sum1_clear    <= 1'b0;
        sum2_clear    <= 1'b0;
        sum3_clear    <= 1'b0;
    end 

    else begin
        case (main_fsm) 

            IDLE: begin

                if (enable == 1'b1) begin
                    main_fsm <= MULT_STATE;
                    mult_enable   <= 1'b1;
                end

                sum1_2_enable <= 1'b0;
                sum3_enable   <= 1'b0;
                sum1_clear    <= 1'b1;
                sum2_clear    <= 1'b1;
                sum3_clear    <= 1'b1;

            end

            MULT_STATE: begin

                if (al_bl_valid == 1'b1 && al_bh_valid == 1'b1 && ah_bl_valid == 1'b1 && ah_bh_valid == 1'b1) begin
                    main_fsm <= SUM_STATE1;
                    sum1_2_enable <= 1'b1;
                end

                mult_enable <= 1'b0;
                sum1_clear  <= 1'b0;
                sum2_clear  <= 1'b0;
                sum3_clear  <= 1'b0;
                
            end

            SUM_STATE1 : begin

                if(sum1_valid == 1'b1 &&  sum2_valid == 1'b1 && sum1_2_enable == 1'b0) begin
                    main_fsm <= SUM_STATE2;
                    sum3_enable <= 1'b1;
                end

                sum1_2_enable <= 1'b0;

            end

            SUM_STATE2 : begin

                if(result_valid == 1'b1  && sum3_enable == 1'b0) begin
                    main_fsm <= RESULT_STATE;
                end

                sum3_enable <= 1'b0;

            end

            RESULT_STATE : begin

                if (enable == 1'b0) begin
                    main_fsm <= IDLE;
                end

            end

            default: begin

                main_fsm    <= IDLE;
                sum3_enable   <= 1'b0;
                sum1_2_enable <= 1'b0;
                sum1_clear    <= 1'b0;
                sum2_clear    <= 1'b0;
                sum3_clear    <= 1'b0;
                mult_enable   <= 1'b0;
                
            end
        endcase
    end
end

endmodule